# -*- coding: utf-8 -*-
#
# Copyright (c) 2019~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from typing import Optional, Iterable, Any

from .utils import inject_by_name
from .ioc import ScopedServiceProvider, LifeTime
from .ioc_service_info import BindedServiceInfo


def _get_keys(target, keys):
    if keys is None:
        keys = []

        try:
            # only hashable() canbe key
            hash(target)
            keys.append(target)
        except TypeError:
            pass

        try:
            name = getattr(target, '__name__')
            keys.append(name)
        except AttributeError:
            pass

    if len(keys) == 0:
        raise RuntimeError('if keys is empty, how do you get it ?')

    return keys



class ServiceProviderDecorator:
    __slots__ = ('_service_provider', 'inject_by')

    def __init__(self, service_provider: ScopedServiceProvider):
        self._service_provider = service_provider
        self.inject_by = None

    def _register_with(self, lifetime, target, wraped_target, keys):
        id = object()
        self._service_provider.register(id, wraped_target, lifetime)
        for k in _get_keys(target, keys):
            self._service_provider.register_bind(k, id)

    def singleton(self, factory=None, *, inject_by = None, keys: Optional[Iterable[Any]] = None):
        '''
        a decorator use for register factory.
        factory which should have signature `(ioc) => any` or `() => any`.

        if factory don't have the signature, you need to set `inject_by`.
        `inject_by` is a function that use to convert `factory` signature to `ioc => any`.

        if `inject_by` is None, we will fallback to use `self.inject_by`;
        if `self.inject_by` also be `None`, mean `factory` do not need to convert.

        you can found some `inject_by_*` from `anyioc.utils`.

        if `keys` is not None, it should be a keys list;
        otherwish we will use `factory` and `factory.__name__` (if exists) as keys.
        '''
        def wrapper(target):
            wrapfunc = inject_by or self.inject_by
            wraped_target = wrapfunc(target) if wrapfunc else target
            self._register_with(LifeTime.singleton, target, wraped_target, keys)
            return target

        return wrapper(factory) if factory else wrapper

    def scoped(self, factory=None, *, inject_by = None, keys: Optional[Iterable[Any]] = None):
        '''
        a decorator use for register factory.
        factory which should have signature `(ioc) => any` or `() => any`.

        if factory don't have the signature, you need to set `inject_by`.
        `inject_by` is a function that use to convert `factory` signature to `ioc => any`.

        if `inject_by` is None, we will fallback to use `self.inject_by`;
        if `self.inject_by` also be `None`, mean `factory` do not need to convert.

        you can found some `inject_by_*` from `anyioc.utils`.

        if `keys` is not None, it should be a keys list;
        otherwish we will use `factory` and `factory.__name__` (if exists) as keys.
        '''
        def wrapper(target):
            wraped_target = inject_by(target) if inject_by else target
            self._register_with(LifeTime.scoped, target, wraped_target, keys)
            return target

        return wrapper(factory) if factory else wrapper

    def transient(self, factory=None, *, inject_by = None, keys: Optional[Iterable[Any]] = None):
        '''
        a decorator use for register factory.
        factory which should have signature `(ioc) => any` or `() => any`.

        if factory don't have the signature, you need to set `inject_by`.
        `inject_by` is a function that use to convert `factory` signature to `ioc => any`.

        if `inject_by` is None, we will fallback to use `self.inject_by`;
        if `self.inject_by` also be `None`, mean `factory` do not need to convert.

        you can found some `inject_by_*` from `anyioc.utils`.

        if `keys` is not None, it should be a keys list;
        otherwish we will use `factory` and `factory.__name__` (if exists) as keys.
        '''
        def wrapper(target):
            wraped_target = inject_by(target) if inject_by else target
            self._register_with(LifeTime.transient, target, wraped_target, keys)
            return target

        return wrapper(factory) if factory else wrapper

    def bind(self, new_key, target_key=None):
        '''
        a decorator use for bind class or function to a alias key.

        if `target_key` is `None`, use `__name__` as `target_key`.
        '''
        def binding(target):
            key = target.__name__ if target_key is None else target_key
            self._service_provider.register_bind(new_key, key)
            return target

        return binding



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from abc import abstractmethod
from typing import Any
from collections import ChainMap
from contextlib import ExitStack

from .err import ServiceNotFoundError
from .symbols import Symbols
from .ioc_resolver import IServiceInfoResolver, ServiceInfoChainResolver
from .ioc_service_info import (
    LifeTime,
    IServiceInfo,
    ServiceInfo,
    ProviderServiceInfo,
    ValueServiceInfo,
    GroupedServiceInfo,
    BindedServiceInfo,
)


class IServiceProvider:
    @abstractmethod
    def __getitem__(self, key):
        raise NotImplementedError

    @abstractmethod
    def get(self, key, d=None) -> Any:
        '''
        get a service by key.
        '''
        raise NotImplementedError

    @abstractmethod
    def scope(self):
        '''
        create a scoped service provider for get scoped services.
        '''
        raise NotImplementedError


class ScopedServiceProvider(IServiceProvider):

    def __init__(self, services: ChainMap):
        super().__init__()
        self._services = services
        self._exit_stack = None
        self._services[Symbols.cache] = ValueServiceInfo({})

    def _get_service_info(self, key) -> IServiceInfo:
        try:
            return self._services[key]
        except KeyError:
            pass
        # load missing resolver and resolve service info.
        resolver: IServiceInfoResolver = self._services[Symbols.missing_resolver].get(self)
        return resolver.get(self, key)

    def __getitem__(self, key):
        service_info = self._get_service_info(key)
        try:
            return service_info.get(self)
        except ServiceNotFoundError as err:
            raise ServiceNotFoundError(key, *err.resolve_chain)

    def get(self, key, d=None) -> Any:
        '''
        get a service by key.
        '''
        try:
            return self[key]
        except ServiceNotFoundError as err:
            if len(err.resolve_chain) == 1:
                return d
            raise

    def enter(self, context):
        '''
        enter the context.

        returns the result of the `context.__enter__()` method.
        '''
        if self._exit_stack is None:
            self._exit_stack = ExitStack()
        return self._exit_stack.enter_context(context)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self._exit_stack is not None:
            self._exit_stack.__exit__(*args)

    def register_service_info(self, key, service_info: IServiceInfo):
        '''
        register a IServiceInfo by key.
        '''
        if not isinstance(service_info, IServiceInfo):
            raise TypeError('service_info must be instance of IServiceInfo.')
        self._services[key] = service_info

    def register(self, key, factory, lifetime):
        '''
        register a service factory by key.
        '''
        return self.register_service_info(key, ServiceInfo(self, key, factory, lifetime))

    def register_singleton(self, key, factory):
        '''
        register a singleton service factory by key.
        '''
        return self.register(key, factory, LifeTime.singleton)

    def register_scoped(self, key, factory):
        '''
        register a scoped service factory by key.
        '''
        return self.register(key, factory, LifeTime.scoped)

    def register_transient(self, key, factory):
        '''
        register a transient service factory by key.
        '''
        return self.register(key, factory, LifeTime.transient)

    def register_value(self, key, value):
        '''
        register a value by key.

        equals `register_transient(key, lambda ioc: value)`
        '''
        return self.register_service_info(key, ValueServiceInfo(value))

    def register_group(self, key, keys: list):
        '''
        register a grouped `key` for get other `keys`.

        the `keys` can be a ref and you can update it later.

        for example:

        ``` py
        provider.register_value('str', 'name')
        provider.register_value('int', 1)
        provider.register_group('any', ['str', 'int'])
        assert provider['any'] == ('name', 1)
        ```

        equals `register_transient(key, lambda ioc: tuple(ioc[k] for k in keys))`
        '''
        return self.register_service_info(key, GroupedServiceInfo(keys))

    def register_bind(self, new_key, target_key):
        '''
        bind `new_key` to `target_key` so
        you can use `new_key` as key to get value from service provider.

        equals `register_transient(new_key, lambda ioc: ioc[target_key])`
        '''
        return self.register_service_info(new_key, BindedServiceInfo(target_key))

    def scope(self):
        '''
        create a scoped service provider.
        '''
        return ScopedServiceProvider(self._services.new_child())

    def decorator(self):
        '''
        get a decorator helper for register items.
        '''
        from .decorate import ServiceProviderDecorator
        return ServiceProviderDecorator(self)


class ServiceProvider(ScopedServiceProvider):
    def __init__(self):
        super().__init__(ChainMap())
        self._services[Symbols.provider] = ProviderServiceInfo()
        self._services[Symbols.provider_root] = ValueServiceInfo(self)
        self._services[Symbols.missing_resolver] = ValueServiceInfo(ServiceInfoChainResolver())
        # service alias
        self._services['ioc'] = self._services[Symbols.provider]
        self._services['provider'] = self._services[Symbols.provider]
        self._services['service_provider'] = self._services[Symbols.provider]



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import sys
from typing import List, Tuple, Union, Any, Dict
from inspect import signature, Parameter

def injectable(*pos_args: List[Union[Tuple[Any], Tuple[Any, Any]]],
               **kw_args: Dict[str, Union[Tuple[Any], Tuple[Any, Any]]]):
    '''
    each **value** in `pos_args` and `kw_args` should be a tuple,
    the 1st item in tuple is the key for get service from ioc container;

    if len of tuple is 2,
    the 2rd item is the default value.

    example:

    ``` py
    @injectable(a=('ioc-key', 'default-value-for-param'))
    def func(a):
        pass
    ```
    '''
    def wrapper(func):
        def new_func(ioc):
            args = []
            for item in pos_args:
                if len(item) == 1:
                    args.append(ioc[item[0]])
                else:
                    key, default = item
                    args.append(ioc.get(key, default))
            kwargs = {}
            for name, item in kw_args.items():
                if len(item) == 1:
                    kwargs[name] = ioc[item[0]]
                else:
                    key, default = item
                    kwargs[name] = ioc.get(key, default)
            return func(*args, **kwargs)
        return new_func
    return wrapper

def inject_by_name(func):
    '''
    wrap the func and auto inject by parameter name.

    var keyword parameter and var positional parameter will not be inject.

    return the new func with signature: `(ioc) => any`
    '''
    sign = signature(func)
    params = [p for p in sign.parameters.values()]
    pos_args = []
    kw_args = {}
    for param in params:
        val = (param.name, ) if param.default is Parameter.empty else (param.name, param.default)
        if param.kind == Parameter.POSITIONAL_OR_KEYWORD:
            pos_args.append(val)
        elif param.kind == Parameter.KEYWORD_ONLY:
            kw_args[param.name] = val
    return injectable(*pos_args, **kw_args)(func)

def inject_by_anno(func):
    '''
    wrap the func and auto inject by parameter annotation.

    var keyword parameter and var positional parameter will not be inject.

    return the new func with signature: `(ioc) => any`
    '''
    sign = signature(func)
    params = [p for p in sign.parameters.values()]
    pos_args = []
    kw_args = {}
    for param in params:
        anno = param.annotation
        if anno is Parameter.empty:
            if param.default is Parameter.empty:
                raise ValueError(f'annotation of args {param.name} is empty.')
            # use `object()` for ensure you never get the value.
            ioc_key = object()
        else:
            ioc_key = anno
        val = (ioc_key, ) if param.default is Parameter.empty else (ioc_key, param.default)

        if param.kind == Parameter.POSITIONAL_OR_KEYWORD:
            pos_args.append(val)
        elif param.kind == Parameter.KEYWORD_ONLY:
            kw_args[param.name] = val

    return injectable(*pos_args, **kw_args)(func)

def inject_by_keys(**keys):
    '''
    wrap the func and auto inject by keys.

    each key in keys should be parameter name of func.

    each value in keys should use for get value from service provider.

    return a decorator for wrap your func to new func with signature: `(ioc) => any`.
    '''

    kw_args = dict((k, (v, )) for k, v in keys.items())
    return injectable(**kw_args)

def auto_enter(func):
    '''
    auto enter the context manager when it created.

    the signature of func should be `(ioc) => any`.
    '''
    def new_func(ioc):
        item = func(ioc)
        ioc.enter(item)
        return item
    return new_func

def dispose_at_exit(provider):
    '''
    register `provider.__exit__()` into `atexit` module.

    return the `provider` itself.
    '''
    import atexit
    @atexit.register
    def provider_dispose_at_exit():
        provider.__exit__(*sys.exc_info())
    return provider

def make_group(container, group_key):
    '''
    add a new group into `container` by `group_key`,
    return a decorator function for add next group item key.

    '''
    group_keys = []
    container.register_group(group_key, group_keys)
    def decorator(next_group_key):
        group_keys.append(next_group_key)
        return next_group_key
    return decorator

# keep old func names:

auto_inject = inject_by_name



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

class Symbols:
    '''
    the symbols use for ioc.

    this keys are predefined in `ServiceProvider`.
    overwrite this keys will break the expected behavior.
    '''

    # current scoped `IServiceProvider`
    provider = object()

    # the root `IServiceProvider`
    provider_root = object()

    # the cache dict from `IServiceProvider`
    cache = object()

    # the missing resolver from `IServiceProvider`
    missing_resolver = object()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from abc import abstractmethod, ABC
from enum import Enum
from inspect import signature, Parameter
from typing import Any
from threading import RLock

from .symbols import Symbols


class LifeTime(Enum):
    transient = 0
    scoped = 1
    singleton = 2


class IServiceInfo(ABC):
    __slots__ = ()

    @abstractmethod
    def get(self, provider) -> Any:
        raise NotImplementedError


class ServiceInfo(IServiceInfo):
    '''generic `IServiceInfo`.'''

    __slots__ = (
        '_key', '_lifetime', '_factory',
        # for not transient
        '_lock',
        # for singleton
        '_cache_value', '_service_provider'
    )

    def __init__(self, service_provider, key, factory, lifetime):

        sign = signature(factory)
        if not sign.parameters:
            self._factory = lambda _: factory()
        elif len(sign.parameters) == 1:
            arg_0 = list(sign.parameters.values())[0]
            if arg_0.kind != Parameter.POSITIONAL_OR_KEYWORD:
                raise TypeError('1st parameter of factory must be a positional parameter.')
            self._factory = factory
        else:
            raise TypeError('factory has too many parameters.')

        self._key = key
        self._lifetime = lifetime
        self._cache_value = None
        self._service_provider = service_provider

        if self._lifetime != LifeTime.transient:
            self._lock = RLock()
        else:
            self._lock = None

        if self._lifetime == LifeTime.singleton:
            # service_provider is required when lifetime == singleton
            assert self._service_provider is not None

    def get(self, provider):
        if self._lifetime is LifeTime.transient:
            return self._factory(provider)

        if self._lifetime is LifeTime.scoped:
            return self._from_scoped(provider)

        if self._lifetime is LifeTime.singleton:
            return self._from_singleton()

        raise NotImplementedError(f'what is {self._lifetime}?')

    def _from_scoped(self, provider):
        cache: dict = provider[Symbols.cache]
        try:
            return cache[self]
        except KeyError:
            pass
        with self._lock:
            try:
                return cache[self]
            except KeyError:
                value = self._factory(provider)
                cache[self] = value
                return value

    def _from_singleton(self):
        if self._cache_value is None:
            with self._lock:
                if self._cache_value is None:
                    self._cache_value = (self._factory(self._service_provider), )
        return self._cache_value[0]


class ProviderServiceInfo(IServiceInfo):
    '''a `IServiceInfo` use for get current `ServiceProvider`.'''

    __slots__ = ()

    def get(self, provider):
        return provider


class ValueServiceInfo(IServiceInfo):
    '''a `IServiceInfo` use for get fixed value.'''

    __slots__ = ('_value')

    def __init__(self, value):
        self._value = value

    def get(self, provider):
        return self._value


class GroupedServiceInfo(IServiceInfo):
    '''a `IServiceInfo` use for get multi values as a tuple from keys list.'''

    __slots__ = ('_keys')

    def __init__(self, keys: list):
        self._keys = keys

    def get(self, provider):
        return tuple(provider[k] for k in self._keys)


class BindedServiceInfo(IServiceInfo):
    '''a `IServiceInfo` use for get value from target key.'''

    __slots__ = ('_target_key')

    def __init__(self, target_key):
        self._target_key = target_key

    def get(self, provider):
        return provider[self._target_key]



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
# a global ioc
# ----------

import functools
import inspect

from .ioc import ServiceProvider
from .utils import inject_by_name, dispose_at_exit

ioc = ServiceProvider()
dispose_at_exit(ioc)

ioc_decorator = ioc.decorator()

ioc_singleton = ioc_decorator.singleton
ioc_scoped = ioc_decorator.scoped
ioc_transient = ioc_decorator.transient
ioc_singleton_cls = functools.partial(ioc_decorator.singleton, inject_by=inject_by_name)
ioc_scoped_cls = functools.partial(ioc_decorator.scoped, inject_by=inject_by_name)
ioc_transient_cls = functools.partial(ioc_decorator.transient, inject_by=inject_by_name)
ioc_bind = ioc_decorator.bind

# scoped global ioc

def _make_module_scoped_provider():
    from .ioc_resolver import IServiceInfoResolver
    from .ioc_service_info import ValueServiceInfo, IServiceInfo
    from .symbols import Symbols

    class ServiceProviderServiceInfoResolver(IServiceInfoResolver):
        def get(self, provider, key) -> IServiceInfo:
            return ValueServiceInfo(ServiceProvider())

    provider = ServiceProvider()
    provider[Symbols.missing_resolver].append(
        ServiceProviderServiceInfoResolver().cache(sync=True)
    )
    dispose_at_exit(provider)
    return provider

_module_scoped_provider = _make_module_scoped_provider()

def get_module_provider(module_name: str=None) -> ServiceProvider:
    '''
    get the module scoped singleton `ServiceProvider`.

    if `module_name` is `None`, use caller module name.
    '''
    if module_name is None:
        fr = inspect.getouterframes(inspect.currentframe())[1]
        mo = inspect.getmodule(fr.frame)
        module_name = mo.__name__

    return _module_scoped_provider[module_name]

def get_package_provider(package_name: str=None) -> ServiceProvider:
    '''
    get the TOP package scoped singleton `ServiceProvider`.

    if `package_name` is `None`, use caller package name.

    for example, `get_package_provider('A.B.C.D')` is equals `get_module_provider('A')`
    '''
    if package_name is None:
        fr = inspect.getouterframes(inspect.currentframe())[1]
        mo = inspect.getmodule(fr.frame)
        package_name = mo.__name__

    package_name = package_name.partition('.')[0]
    return _module_scoped_provider[package_name]
